pub mod error;

use error::{handle_rejection, InternalServerError, NotFoundError};
use hyper::{body, Body};
use std::path::PathBuf;

use bytes::Bytes;
use clap::Parser as _;
use tokio::io::AsyncReadExt;
use tracing::{debug, warn, Level};
use warp::{
    filters::{any::any, path::param},
    http::HeaderMap,
    hyper,
    reject::Rejection,
    Filter,
};

#[derive(clap::Parser)]
pub struct Args {
    music_dir: PathBuf,
}

#[tokio::main]
async fn main() {
    let config: &'static Args = Box::leak(Box::new(Args::parse()));

    tracing_subscriber::fmt().with_max_level(Level::INFO).init();

    let route = warp::get()
        .and(warp::path("track"))
        .and(any().map(move || config))
        .and(warp::header::headers_cloned())
        .and(param())
        .and_then(serve_local_tracks);

    warp::serve(route.recover(handle_rejection))
        .run(([127, 0, 0, 1], 8080))
        .await;
}

pub async fn serve_local_tracks(
    config: &Args,
    _headers: HeaderMap,
    // TODO: request files based off of their MusicBrainz Identifier instead
    track_name: String,
) -> Result<warp::reply::Response, Rejection> {
    let mut location = config.music_dir.clone();
    // FIXME: for now, file paths need to be URL safe to be able to be requested
    location.push(PathBuf::from(track_name));

    let Ok(file) = tokio::fs::File::options().read(true).open(&location).await else {
        return Err(NotFoundError(
            format!("The requested song could not be found on disk. Tried loading {location:?}")
                .into(),
        )
        .into());
    };

    // TODO: handle range requests

    let (sender, body) = Body::channel();

    debug!("Starting to stream file");
    tokio::task::spawn(stream_file(sender, file));

    if let Ok(response) = hyper::Response::builder()
        .header("Content-Type", "audio/mpeg")
        // .header("Accept-Ranges", "bytes") // TODO: handle range requests
        .body(body)
    {
        Ok(response)
    } else {
        Err(InternalServerError("Failed to build Response".into()).into())
    }
}

pub async fn stream_file(mut dest: body::Sender, mut file: tokio::fs::File) {
    let mut buf = [0u8; 512];

    loop {
        let bytes_read = match file.read(&mut buf).await {
            Ok(0) => break debug!("Done streaming a file"),
            Ok(bytes_read) => bytes_read,
            Err(err) => break warn!("Couldn't read part of a file. Got err: {err}"),
        };

        let res = dest
            .send_data(Bytes::copy_from_slice(&buf[..bytes_read]))
            .await;

        if res.is_err() {
            warn!("Failed to send some data to a client");
            break;
        }
    }
}
